package com.sun.xml.internal.messaging.saaj.soap.impl;

import com.sun.xml.internal.messaging.saaj.util.FastInfosetReflection;
import javax.xml.transform.Transformer;
import java.io.IOException;
import java.io.OutputStreamWriter;
import javax.xml.transform.stream.StreamResult;
import com.sun.xml.internal.messaging.saaj.util.transform.EfficientStreamingTransformer;
import java.io.OutputStream;
import java.util.logging.Level;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.Source;
import javax.xml.soap.SOAPBody;
import java.util.Iterator;
import com.sun.xml.internal.messaging.saaj.SOAPExceptionImpl;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPException;
import com.sun.xml.internal.messaging.saaj.soap.name.NameImpl;
import javax.xml.namespace.QName;
import javax.xml.soap.Name;
import com.sun.xml.internal.messaging.saaj.soap.SOAPDocumentImpl;
import com.sun.xml.internal.messaging.saaj.soap.Envelope;

public abstract class EnvelopeImpl extends ElementImpl implements Envelope
{
    protected HeaderImpl header;
    protected BodyImpl body;
    String omitXmlDecl;
    String charset;
    String xmlDecl;
    
    protected EnvelopeImpl(SOAPDocumentImpl ownerDoc, Name name) {
        super(ownerDoc, name);
        this.omitXmlDecl = "yes";
        this.charset = "utf-8";
        this.xmlDecl = null;
    }
    
    protected EnvelopeImpl(SOAPDocumentImpl ownerDoc, QName name) {
        super(ownerDoc, name);
        this.omitXmlDecl = "yes";
        this.charset = "utf-8";
        this.xmlDecl = null;
    }
    
    protected EnvelopeImpl(SOAPDocumentImpl ownerDoc, NameImpl name, boolean createHeader, boolean createBody) throws SOAPException {
        this(ownerDoc, name);
        this.ensureNamespaceIsDeclared(this.getElementQName().getPrefix(), this.getElementQName().getNamespaceURI());
        if (createHeader) {
            this.addHeader();
        }
        if (createBody) {
            this.addBody();
        }
    }
    
    protected abstract NameImpl getHeaderName(String p0);
    
    protected abstract NameImpl getBodyName(String p0);
    
    public SOAPHeader addHeader() throws SOAPException {
        return this.addHeader(null);
    }
    
    public SOAPHeader addHeader(String prefix) throws SOAPException {
        if (prefix == null || prefix.equals("")) {
            prefix = this.getPrefix();
        }
        NameImpl headerName = this.getHeaderName(prefix);
        NameImpl bodyName = this.getBodyName(prefix);
        HeaderImpl header = null;
        SOAPElement firstChild = null;
        Iterator<Object> eachChild = this.getChildElementNodes();
        if (eachChild.hasNext()) {
            firstChild = (SOAPElement)eachChild.next();
            if (firstChild.getElementName().equals(headerName)) {
                EnvelopeImpl.log.severe("SAAJ0120.impl.header.already.exists");
                throw new SOAPExceptionImpl("Can't add a header when one is already present.");
            }
            if (!firstChild.getElementName().equals(bodyName)) {
                EnvelopeImpl.log.severe("SAAJ0121.impl.invalid.first.child.of.envelope");
                throw new SOAPExceptionImpl("First child of Envelope must be either a Header or Body");
            }
        }
        header = (HeaderImpl)this.createElement(headerName);
        this.insertBefore(header, firstChild);
        header.ensureNamespaceIsDeclared(headerName.getPrefix(), headerName.getURI());
        return header;
    }
    
    protected void lookForHeader() throws SOAPException {
        NameImpl headerName = this.getHeaderName(null);
        HeaderImpl hdr = (HeaderImpl)this.findChild(headerName);
        this.header = hdr;
    }
    
    public SOAPHeader getHeader() throws SOAPException {
        this.lookForHeader();
        return this.header;
    }
    
    protected void lookForBody() throws SOAPException {
        NameImpl bodyName = this.getBodyName(null);
        BodyImpl bodyChildElement = (BodyImpl)this.findChild(bodyName);
        this.body = bodyChildElement;
    }
    
    public SOAPBody addBody() throws SOAPException {
        return this.addBody(null);
    }
    
    public SOAPBody addBody(String prefix) throws SOAPException {
        this.lookForBody();
        if (prefix == null || prefix.equals("")) {
            prefix = this.getPrefix();
        }
        if (this.body == null) {
            NameImpl bodyName = this.getBodyName(prefix);
            this.insertBefore(this.body = (BodyImpl)this.createElement(bodyName), null);
            this.body.ensureNamespaceIsDeclared(bodyName.getPrefix(), bodyName.getURI());
            return this.body;
        }
        EnvelopeImpl.log.severe("SAAJ0122.impl.body.already.exists");
        throw new SOAPExceptionImpl("Can't add a body when one is already present.");
    }
    
    @Override
    protected SOAPElement addElement(Name name) throws SOAPException {
        if (this.getBodyName(null).equals(name)) {
            return this.addBody(name.getPrefix());
        }
        if (this.getHeaderName(null).equals(name)) {
            return this.addHeader(name.getPrefix());
        }
        return super.addElement(name);
    }
    
    @Override
    protected SOAPElement addElement(QName name) throws SOAPException {
        if (this.getBodyName(null).equals(NameImpl.convertToName(name))) {
            return this.addBody(name.getPrefix());
        }
        if (this.getHeaderName(null).equals(NameImpl.convertToName(name))) {
            return this.addHeader(name.getPrefix());
        }
        return super.addElement(name);
    }
    
    public SOAPBody getBody() throws SOAPException {
        this.lookForBody();
        return this.body;
    }
    
    public Source getContent() {
        return new DOMSource(this.getOwnerDocument());
    }
    
    public Name createName(String localName, String prefix, String uri) throws SOAPException {
        if ("xmlns".equals(prefix)) {
            EnvelopeImpl.log.severe("SAAJ0123.impl.no.reserved.xmlns");
            throw new SOAPExceptionImpl("Cannot declare reserved xmlns prefix");
        }
        if (prefix == null && "xmlns".equals(localName)) {
            EnvelopeImpl.log.severe("SAAJ0124.impl.qualified.name.cannot.be.xmlns");
            throw new SOAPExceptionImpl("Qualified name cannot be xmlns");
        }
        return NameImpl.create(localName, prefix, uri);
    }
    
    public Name createName(String localName, String prefix) throws SOAPException {
        String namespace = this.getNamespaceURI(prefix);
        if (namespace == null) {
            EnvelopeImpl.log.log(Level.SEVERE, "SAAJ0126.impl.cannot.locate.ns", new String[] { prefix });
            throw new SOAPExceptionImpl("Unable to locate namespace for prefix " + prefix);
        }
        return NameImpl.create(localName, prefix, namespace);
    }
    
    public Name createName(String localName) throws SOAPException {
        return NameImpl.createFromUnqualifiedName(localName);
    }
    
    public void setOmitXmlDecl(String value) {
        this.omitXmlDecl = value;
    }
    
    public void setXmlDecl(String value) {
        this.xmlDecl = value;
    }
    
    private String getOmitXmlDecl() {
        return this.omitXmlDecl;
    }
    
    public void setCharsetEncoding(String value) {
        this.charset = value;
    }
    
    public void output(OutputStream out) throws IOException {
        try {
            Transformer transformer = EfficientStreamingTransformer.newTransformer();
            transformer.setOutputProperty("omit-xml-declaration", "yes");
            transformer.setOutputProperty("encoding", this.charset);
            if (this.omitXmlDecl.equals("no") && this.xmlDecl == null) {
                this.xmlDecl = "<?xml version=\"" + this.getOwnerDocument().getXmlVersion() + "\" encoding=\"" + this.charset + "\" ?>";
            }
            StreamResult result = new StreamResult(out);
            if (this.xmlDecl != null) {
                OutputStreamWriter writer = new OutputStreamWriter(out, this.charset);
                writer.write(this.xmlDecl);
                writer.flush();
                result = new StreamResult(writer);
            }
            if (EnvelopeImpl.log.isLoggable(Level.FINE)) {
                EnvelopeImpl.log.log(Level.FINE, "SAAJ0190.impl.set.xml.declaration", new String[] { this.omitXmlDecl });
                EnvelopeImpl.log.log(Level.FINE, "SAAJ0191.impl.set.encoding", new String[] { this.charset });
            }
            transformer.transform(this.getContent(), result);
        }
        catch (Exception ex) {
            throw new IOException(ex.getMessage());
        }
    }
    
    public void output(OutputStream out, boolean isFastInfoset) throws IOException {
        if (!isFastInfoset) {
            this.output(out);
        }
        else {
            try {
                Source source = this.getContent();
                Transformer transformer = EfficientStreamingTransformer.newTransformer();
                transformer.transform(this.getContent(), FastInfosetReflection.FastInfosetResult_new(out));
            }
            catch (Exception ex) {
                throw new IOException(ex.getMessage());
            }
        }
    }
    
    @Override
    public SOAPElement setElementQName(QName newName) throws SOAPException {
        EnvelopeImpl.log.log(Level.SEVERE, "SAAJ0146.impl.invalid.name.change.requested", new Object[] { this.elementQName.getLocalPart(), newName.getLocalPart() });
        throw new SOAPException("Cannot change name for " + this.elementQName.getLocalPart() + " to " + newName.getLocalPart());
    }
}
